/*
  cadaverserver
  Copyright (C) 1999  Erik Thiele

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "zeit.hpp"
#include "fatal.hpp"
#include "myspielfeld.hpp"
#include "conn.hpp"
#include "mycyborg.hpp"
#include "misc.hpp"
#include "config.h"
#include "memchk.hpp"
#include "serverglobals.hpp"
#include "keyboard.hpp"
#include "selector.hpp"

#include <stdio.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/socket.h>
#include <signal.h>

#ifndef YYY_HAVE_SOCKLEN_T
#define socklen_t int
#endif

void usage()
{
  fprintf(stderr,
	  "\n"
	  "\n"
	  "Usage:\n"
	  "cadaverserver xsize ysize\n"
	  "\n"
	  "\n");
  exit(1);
};

int main(int argc, char *argv[])
{
  memchk::resetsituation();
  {

    printf("\nCadaver Server Version %s\n\n",VERSION);
    
    signal (SIGPIPE,SIG_IGN);
    
    srand(time(0));

    viewvisimapper = new visimapper(4.5);
    
    int wantedxsize,wantedysize;
    if (argc!=3) usage();
    if (sscanf(argv[1],"%d",&wantedxsize)!=1) usage();
    if (sscanf(argv[2],"%d",&wantedysize)!=1) usage();
    if ((wantedxsize<=0)||(wantedysize<=0)
	||(wantedxsize>500)||(wantedysize>500)) {
      fprintf(stderr,"invalid sizes.  0<xsize<=500  0<ysize<=500\n");
      exit(1);
    };
    gsf=new spielfeld(wantedxsize,wantedysize);
    gsc=new spielcyborg;
    gcon=new connections;
    randomize_spielfeld();
    
    int sock;
    struct sockaddr_in name;
    if ((sock = socket (PF_INET, SOCK_STREAM, 0))<0) fatal_errno("socket");
    
    int reus=1;
    if (setsockopt(sock,SOL_SOCKET,SO_REUSEADDR,(const char*)&reus,sizeof reus)<0)
      fatal_errno("setsockopt");
    
    name.sin_family = AF_INET;
    name.sin_port = htons (8932);
    name.sin_addr.s_addr = htonl (INADDR_ANY);
    if (bind (sock, (struct sockaddr *) &name, sizeof (name)) < 0) fatal_errno("bind");
    if (listen(sock,5)<0) fatal_errno("listen");
    
    keyboardparser kbdparse;

    fcntl (0,F_SETFL,O_NONBLOCK);
    printf("> "); fflush(stdout);
    
    zeit timeout, interval(0,200000);
    zeit lastcalculation;
    lastcalculation.getcurrenttime();

    selector mysel;

    bool ende=false;
    while (!ende) {
      {
	zeit now;
	timeout=lastcalculation;
	timeout.add(interval);
	now.getcurrenttime();
	timeout.subtract(now);
	if (timeout.tofloat()<0) timeout=zeit(); // timeout NOW !
      };

      mysel.reset();
      mysel.add_read(0);
      mysel.add_read(sock);
      for (map<const int,conn *const>::iterator i=gcon->all.begin();i!=gcon->all.end();i++) {
	mysel.add_read(i->second->gethandle());
	if (i->second->outputbuffer.size()) mysel.add_write(i->second->gethandle());
      };
      mysel.settimeout(timeout);

      mysel.waitnow();

      if (mysel.isready_read(0)) if (!kbdparse.parse()) ende=true;
      if (mysel.isready_read(sock)) {
	  struct sockaddr_in addr;
	  socklen_t addr_size=sizeof(addr);
	  int neuhandle = accept (sock,(struct sockaddr *) &addr, &addr_size);
	  if (neuhandle<0) fatal_errno("accept failed");
	  fcntl (neuhandle,F_SETFL,O_NONBLOCK);
	  conn *x=new conn(neuhandle,addr);
	  gcon->add(x);
      };
      for (map<const int, conn *const>::iterator i=gcon->all.begin();i!=gcon->all.end();) {
	bool term=false;
	int id=i->second->id();
	if (mysel.isready_write(i->second->gethandle())) if (!i->second->manageoutput()) term=true;
	if (mysel.isready_read(i->second->gethandle())) if (!i->second->manageinput()) term=true;
	i++;
	if (term) gcon->del(id);
      };
      if (mysel.istimeoutexpired()) {
	zeit now;
	now.getcurrenttime();
	timeout=now;
	timeout.subtract(lastcalculation);
	float ti=timeout.tofloat();
	calculate_spielfeld(ti);
	calculate_spielcyborg(ti);
	lastcalculation=now;
      };
    };
    
    close (sock);
    
    delete gcon;
    delete gsc;
    delete gsf;
    
    delete viewvisimapper;
  };
  if (!memchk::everythingzero()) {
    cerr << "Memory statistics error !!!" << endl;
    memchk::printstatistics(cerr);
  };
  
  printf("\n\nCadaver server terminating normally\n\n");
  
  return 0;
};
